@@ -7,7 +7,7 @@ module ParseHelper
   require 'parser/macruby'
   require 'parser/rubymotion'
 
-  ALL_VERSIONS = %w(1.8 1.9 2.0 2.1 2.2 2.3 2.4 2.5 2.6 2.7 2.8 mac ios)
+  ALL_VERSIONS = %w(1.8 1.9 2.0 2.1 2.2 2.3 2.4 2.5 2.6 2.7 3.0 3.1 mac ios)
 
   def setup
     @diagnostics = []
@@ -27,7 +27,8 @@ module ParseHelper
     when '2.5' then parser = Parser::Ruby25.new
     when '2.6' then parser = Parser::Ruby26.new
     when '2.7' then parser = Parser::Ruby27.new
-    when '2.8' then parser = Parser::Ruby28.new
+    when '3.0' then parser = Parser::Ruby30.new
+    when '3.1' then parser = Parser::Ruby31.new
     when 'mac' then parser = Parser::MacRuby.new
     when 'ios' then parser = Parser::RubyMotion.new
     else raise "Unrecognized Ruby version #{version}"
@@ -49,15 +50,17 @@ module ParseHelper
     end
   end
 
-  def assert_source_range(begin_pos, end_pos, range, version, what)
-    assert range.is_a?(Parser::Source::Range),
-           "(#{version}) #{range.inspect}.is_a?(Source::Range) for #{what}"
-
-    assert_equal begin_pos, range.begin_pos,
-                 "(#{version}) begin of #{what}"
-
-    assert_equal end_pos, range.end_pos,
-                 "(#{version}) end of #{what}"
+  def assert_source_range(expect_range, range, version, what)
+    if expect_range == nil
+      # Avoid "Use assert_nil if expecting nil from .... This will fail in Minitest 6.""
+      assert_nil range,
+                   "(#{version}) range of #{what}"
+    else
+      assert range.is_a?(Parser::Source::Range),
+             "(#{version}) #{range.inspect}.is_a?(Source::Range) for #{what}"
+      assert_equal expect_range, range.to_range,
+                   "(#{version}) range of #{what}"
+    end
   end
 
   # Use like this:
@@ -82,6 +85,24 @@ module ParseHelper
       parser.instance_eval { @lexer.force_utf32 = true }
       try_parsing(ast, code, parser, source_maps, version)
     end
+
+    # Also check that it doesn't throw anything
+    # except (possibly) Parser::SyntaxError on other versions of Ruby
+    with_versions(ALL_VERSIONS - versions) do |version, parser|
+      begin
+        source_file = Parser::Source::Buffer.new('(assert_older_rubies)', source: code)
+        parser.parse(source_file)
+      rescue Parser::SyntaxError
+        # ok
+      rescue StandardError
+        # unacceptable
+        raise
+      else
+        # No error means that `code` is valid for `version`, but has a different meaning.
+        # Sometimes Ruby has breaking changes (like numparams)
+        # that re-use constructions from previous versions.
+      end
+    end
   end
 
   def try_parsing(ast, code, parser, source_maps, version)
@@ -105,7 +126,7 @@ module ParseHelper
     assert_equal ast, parsed_ast,
                  "(#{version}) AST equality"
 
-    parse_source_map_descriptions(source_maps) do |begin_pos, end_pos, map_field, ast_path, line|
+    parse_source_map_descriptions(source_maps) do |range, map_field, ast_path, line|
 
       astlet = traverse_ast(parsed_ast, ast_path)
 
@@ -119,16 +140,12 @@ module ParseHelper
       assert astlet.location.respond_to?(map_field),
              "(#{version}) #{astlet.location.inspect}.respond_to?(#{map_field.inspect}) for:\n#{parsed_ast.inspect}"
 
-      range = astlet.location.send(map_field)
+      found_range = astlet.location.send(map_field)
 
-      assert_source_range(begin_pos, end_pos, range, version, line.inspect)
+      assert_source_range(range, found_range, version, line.inspect)
     end
 
-    assert parser.instance_eval { @lexer }.cmdarg.empty?,
-      "(#{version}) expected cmdarg to be empty after parsing"
-
-    assert_equal 0, parser.instance_eval { @lexer.instance_eval { @paren_nest } },
-      "(#{version}) expected paren_nest to be 0 after parsing"
+    assert_state_is_final(parser, version)
   end
 
   # Use like this:
@@ -164,18 +181,18 @@ module ParseHelper
       assert_equal arguments, emitted_diagnostic.arguments
       assert_equal message, emitted_diagnostic.message
 
-      parse_source_map_descriptions(source_maps) do |begin_pos, end_pos, map_field, ast_path, line|
+      parse_source_map_descriptions(source_maps) do |range, map_field, ast_path, line|
 
         case map_field
         when 'location'
-          assert_source_range begin_pos, end_pos,
+          assert_source_range range,
                               emitted_diagnostic.location,
                               version, 'location'
 
         when 'highlights'
           index = ast_path.first.to_i
 
-          assert_source_range begin_pos, end_pos,
+          assert_source_range range,
                               emitted_diagnostic.highlights[index],
                               version, "#{index}th highlight"
 
@@ -254,7 +271,7 @@ module ParseHelper
   SOURCE_MAP_DESCRIPTION_RE =
       /(?x)
        ^(?# $1 skip)            ^(\s*)
-        (?# $2 highlight)        ([~\^]+)
+        (?# $2 highlight)        ([~\^]+|\!)
                                  \s+
         (?# $3 source_map_field) ([a-z_]+)
         (?# $5 ast_path)         (\s+\(([a-z_.\/0-9]+)\))?
@@ -272,8 +289,11 @@ module ParseHelper
       next if line.empty?
 
       if (match = SOURCE_MAP_DESCRIPTION_RE.match(line))
-        begin_pos        = match[1].length
-        end_pos          = begin_pos + match[2].length
+        if match[2] != '!'
+          begin_pos        = match[1].length
+          end_pos          = begin_pos + match[2].length
+          range            = begin_pos...end_pos
+        end
         source_map_field = match[3]
 
         if match[5]
@@ -282,7 +302,7 @@ module ParseHelper
           ast_path = []
         end
 
-        yield begin_pos, end_pos, source_map_field, ast_path, line
+        yield range, source_map_field, ast_path, line
       else
         raise "Cannot parse source map description line: #{line.inspect}."
       end
@@ -320,4 +340,24 @@ module ParseHelper
 
     result
   end
+
+  def assert_state_is_final(parser, version)
+    lexer = parser.lexer
+
+    assert lexer.cmdarg.empty?, "(#{version}) expected cmdarg to be empty after parsing"
+    assert lexer.cond.empty?, "(#{version}) expected cond to be empty after parsing"
+
+    assert lexer.cmdarg_stack.empty?, "(#{version}) expected cmdarg_stack to be empty after parsing"
+    assert lexer.cond_stack.empty?, "(#{version}) expected cond_stack to be empty after parsing"
+
+    assert_equal 0, lexer.paren_nest, "(#{version}) expected paren_nest to be 0 after parsing"
+    assert lexer.lambda_stack.empty?, "(#{version}) expected lambda_stack to be empty after parsing"
+
+    assert parser.static_env.empty?, "(#{version}) expected static_env to be empty after parsing"
+    assert parser.context.empty?, "(#{version}) expected context to be empty after parsing"
+    assert parser.max_numparam_stack.empty?, "(#{version}) expected max_numparam_stack to be empty after parsing"
+    assert parser.current_arg_stack.empty?, "(#{version}) expected current_arg_stack to be empty after parsing"
+    assert parser.pattern_variables.empty?, "(#{version}) expected pattern_variables to be empty after parsing"
+    assert parser.pattern_hash_keys.empty?, "(#{version}) expected pattern_hash_keys to be empty after parsing"
+  end
 end
